from typing import Optional, Any, Callable, List, Dict, TYPE_CHECKING
import os
import sys
import tomli
from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json
from importlib import util
from chainlit.logger import logger
from chainlit.version import __version__

if TYPE_CHECKING:
    from chainlit.action import Action

PACKAGE_ROOT = os.path.dirname(__file__)

# Get the directory the script is running from
APP_ROOT = os.getcwd()

config_dir = os.path.join(APP_ROOT, ".chainlit")
config_file = os.path.join(config_dir, "config.toml")

# Default config file created if none exists
DEFAULT_CONFIG_STR = f"""[project]
# If true (default), the app will be available to anonymous users.
# If false, users will need to authenticate and be part of the project to use the app.
public = true

# The project ID (found on https://cloud.chainlit.io).
# If provided, all the message data will be stored in the cloud.
# The project ID is required when public is set to false.
#id = ""

# Whether to enable telemetry (default: true). No personal data is collected.
enable_telemetry = true

# List of environment variables to be provided by each user to use the app.
user_env = []

[UI]
# Name of the app and chatbot.
name = "Chatbot"

# Description of the app and chatbot. This is used for HTML tags.
# description = ""

# The default value for the expand messages settings.
default_expand_messages = false

# Hide the chain of thought details from the user in the UI.
hide_cot = false

# Link to your github repo. This will add a github button in the UI's header.
# github = ""

[meta]
generated_by = "{__version__}"
"""

chainlit_prod_url = os.environ.get("CHAINLIT_PROD_URL")
chainlit_server = "https://cloud.chainlit.io"


DEFAULT_HOST = "0.0.0.0"
DEFAULT_PORT = 8000


@dataclass()
class RunSettings:
    # Name of the module (python file) used in the run command
    module_name: Optional[str] = None
    host: str = DEFAULT_HOST
    port: int = DEFAULT_PORT
    headless: bool = False
    watch: bool = False
    no_cache: bool = False
    debug: bool = False
    ci: bool = False


@dataclass_json
@dataclass()
class UISettings:
    name: str
    description: str = ""
    hide_cot: bool = False
    default_expand_messages: bool = False
    github: str = None


@dataclass()
class CodeSettings:
    # Developer defined callbacks for each action. Key is the action name, value is the callback function.
    action_callbacks: Dict[str, Callable[["Action"], Any]]
    # Module object loaded from the module_name
    module: Any = None
    # Bunch of callbacks defined by the developer
    on_stop: Optional[Callable[[], Any]] = None
    on_chat_start: Optional[Callable[[], Any]] = None
    on_message: Optional[Callable[[str], Any]] = None
    lc_agent_is_async: Optional[bool] = None
    lc_run: Optional[Callable[[Any, str], str]] = None
    lc_postprocess: Optional[Callable[[Any], str]] = None
    lc_factory: Optional[Callable[[], Any]] = None
    lc_rename: Optional[Callable[[str], str]] = None

    def validate(self):
        requires_one_of = [
            "lc_factory",
            "on_message",
            "on_chat_start",
        ]

        mutually_exclusive = ["lc_factory"]

        # Check if at least one of the required attributes is set
        if not any(getattr(self, attr) for attr in requires_one_of):
            raise ValueError(
                f"Module should at least expose one of {', '.join(requires_one_of)} function"
            )

        # Check if any mutually exclusive attributes are set together
        for i, attr1 in enumerate(mutually_exclusive):
            for attr2 in mutually_exclusive[i + 1 :]:
                if getattr(self, attr1) and getattr(self, attr2):
                    raise ValueError(
                        f"Module should not expose both {attr1} and {attr2} functions"
                    )

        return True


@dataclass_json
@dataclass()
class ProjectSettings:
    # Enables Cloud features if provided
    id: Optional[str] = None
    # Whether the app is available to anonymous users or only to team members.
    public: bool = True
    # Whether to enable telemetry. No personal data is collected.
    enable_telemetry: bool = True
    # List of environment variables to be provided by each user to use the app. If empty, no environment variables will be asked to the user.
    user_env: List[str] = None
    # Path to the local langchain cache database
    lc_cache_path: str = None


@dataclass()
class ChainlitConfig:
    # Directory where the Chainlit project is located
    root = APP_ROOT
    # Chainlit server URL. Used only for cloud features
    chainlit_server: str
    # The url of the deployed app. Only set if the app is deployed.
    chainlit_prod_url = chainlit_prod_url

    run: RunSettings
    ui: UISettings
    project: ProjectSettings
    code: CodeSettings


def init_config(log=False):
    """Initialize the configuration file if it doesn't exist."""
    if not os.path.exists(config_file):
        os.makedirs(config_dir, exist_ok=True)
        with open(config_file, "w", encoding="utf-8") as f:
            f.write(DEFAULT_CONFIG_STR)
            logger.info(f"Created default config file at {config_file}")
    elif log:
        logger.info(f"Config file already exists at {config_file}")


def load_module(target: str):
    """Load the specified module."""

    # Get the target's directory
    target_dir = os.path.dirname(os.path.abspath(target))

    # Add the target's directory to the Python path
    sys.path.insert(0, target_dir)

    spec = util.spec_from_file_location(target, target)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Remove the target's directory from the Python path
    sys.path.pop(0)

    config.code.validate()


def load_settings():
    with open(config_file, "rb") as f:
        toml_dict = tomli.load(f)
        # Load project settings
        project_config = toml_dict.get("project", {})
        ui_settings = toml_dict.get("UI", {})
        meta = toml_dict.get("meta")

        if not meta or meta.get("generated_by") <= "0.3.0":
            raise ValueError(
                "Your config file is outdated. Please delete it and restart the app to regenerate it."
            )

        lc_cache_path = os.path.join(config_dir, ".langchain.db")

        project_settings = ProjectSettings(
            lc_cache_path=lc_cache_path, **project_config
        )
        ui_settings = UISettings(**ui_settings)

        if not project_settings.public and not project_settings.id:
            raise ValueError("Project ID is required when public is set to false.")

        return {
            "ui": ui_settings,
            "project": project_settings,
            "code": CodeSettings(action_callbacks={}),
        }


def reload_config():
    """Reload the configuration from the config file."""
    global config
    if config is None:
        return

    settings = load_settings()

    config.code = settings["code"]
    config.ui = settings["ui"]
    config.project = settings["project"]


def load_config():
    """Load the configuration from the config file."""
    init_config()

    settings = load_settings()

    config = ChainlitConfig(
        chainlit_server=chainlit_server,
        chainlit_prod_url=chainlit_prod_url,
        run=RunSettings(),
        **settings,
    )

    return config


config = load_config()



from typing import Dict
from chainlit.emitter import get_emitter

user_sessions: Dict[str, Dict] = {}


class UserSession:
    """
    Developer facing user session class.
    Useful for the developer to store user specific data between calls.
    """

    def get(self, key):
        emitter = get_emitter()
        if not emitter:
            return None

        if emitter.session["id"] not in user_sessions:
            # Create a new user session
            user_sessions[emitter.session["id"]] = {}

        user_session = user_sessions[emitter.session["id"]]

        # Copy important fields from the session
        user_session["id"] = emitter.session["id"]
        user_session["env"] = emitter.session["user_env"]
        if "agent" in emitter.session:
            user_session["agent"] = emitter.session["agent"]

        return user_session.get(key)

    def set(self, key, value):
        emitter = get_emitter()
        if not emitter:
            return None

        if emitter.session["id"] not in user_sessions:
            user_sessions[emitter.session["id"]] = {}

        user_session = user_sessions[emitter.session["id"]]
        user_session[key] = value


user_session = UserSession()



from typing import Any, Callable

import asyncio
from syncer import sync
from asyncer import asyncify

from chainlit.emitter import get_emitter


def make_async(function: Callable):
    emitter = get_emitter()
    if not emitter:
        raise RuntimeError(
            "Emitter not found, please call make_async in a Chainlit context."
        )

    def wrapper(*args, **kwargs):
        emitter.session["running_sync"] = True
        __chainlit_emitter__ = emitter
        res = function(*args, **kwargs)
        emitter.session["running_sync"] = False
        return res

    return asyncify(wrapper, cancellable=True)


def run_sync(co: Any):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError as e:
        if "There is no current event loop" in str(e):
            loop = None

    if loop is None or not loop.is_running():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return sync(co)



from typing import List, Dict, Union
from abc import ABC, abstractmethod
import uuid
import time
import asyncio

from chainlit.telemetry import trace_event
from chainlit.emitter import get_emitter
from chainlit.config import config
from chainlit.types import (
    LLMSettings,
    AskSpec,
    AskFileSpec,
    AskFileResponse,
    AskResponse,
)
from chainlit.element import Element
from chainlit.action import Action


def current_milli_time():
    """Get the current time in milliseconds."""
    return round(time.time() * 1000)


class MessageBase(ABC):
    id: int = None
    temp_id: str = None
    streaming = False
    created_at: int = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.temp_id = uuid.uuid4().hex
        self.created_at = current_milli_time()
        self.emitter = get_emitter()
        if not self.emitter:
            raise RuntimeError("Message should be instantiated in a Chainlit context")

    @abstractmethod
    def to_dict(self):
        pass

    async def _create(self):
        msg_dict = self.to_dict()
        if self.emitter.client and not self.id:
            self.id = await self.emitter.client.create_message(msg_dict)
            if self.id:
                msg_dict["id"] = self.id

        return msg_dict

    async def update(
        self,
        author: str = None,
        content: str = None,
        language: str = None,
        prompt: str = None,
        llm_settings: LLMSettings = None,
    ):
        """
        Update a message already sent to the UI.
        """
        trace_event("update_message")

        if author:
            self.author = author
        if content:
            self.content = content
        if language:
            self.language = language
        if prompt:
            self.prompt = prompt
        if llm_settings:
            self.llmSettings = llm_settings

        msg_dict = self.to_dict()

        if self.emitter.client and self.id:
            self.emitter.client.update_message(self.id, msg_dict)
            msg_dict["id"] = self.id

        await self.emitter.update_message(msg_dict)

        return True

    async def remove(self):
        """
        Remove a message already sent to the UI.
        This will not automatically remove potential nested messages and could lead to undesirable side effects in the UI.
        """
        trace_event("remove_message")

        if self.emitter.client and self.id:
            await self.emitter.client.delete_message(self.id)

        await self.emitter.delete_message(self.to_dict())

        return True

    async def send(self) -> Union[str, int]:
        if self.content is None:
            self.content = ""

        msg_dict = await self._create()

        if self.streaming:
            self.streaming = False

        await self.emitter.send_message(msg_dict)

        return self.id or self.temp_id

    async def stream_token(self, token: str):
        """
        Sends a token to the UI. This is useful for streaming messages.
        Once all tokens have been streamed, call .send() to persist the message.
        """

        if not self.streaming:
            self.streaming = True
            msg_dict = self.to_dict()
            await self.emitter.stream_start(msg_dict)

        self.content += token
        await self.emitter.send_token(id=self.id or self.temp_id, token=token)


class Message(MessageBase):
    """
    Send a message to the UI
    If a project ID is configured, the message will be persisted in the cloud.

    Args:
        content (str): The content of the message.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        prompt (str, optional): The prompt used to generate the message. If provided, enables the prompt playground for this message.
        llm_settings (LLMSettings, optional): Settings of the LLM used to generate the prompt. This is useful for debug purposes in the prompt playground.
        language (str, optional): Language of the code is the content is code. See https://react-code-blocks-rajinwonderland.vercel.app/?path=/story/codeblock--supported-languages for a list of supported languages.
        indent (int, optional): If positive, the message will be nested in the UI.
        actions (List[Action], optional): A list of actions to send with the message.
        elements (List[Element], optional): A list of elements to send with the message.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        prompt: str = None,
        llm_settings: LLMSettings = None,
        language: str = None,
        indent: int = 0,
        actions: List[Action] = [],
        elements: List[Element] = [],
    ):
        self.content = content
        self.author = author
        self.prompt = prompt
        self.language = language
        self.indent = indent
        self.actions = actions
        self.elements = elements
        self.llmSettings = None

        if llm_settings is None and prompt is not None:
            self.llmSettings = LLMSettings().to_dict()

        if llm_settings:
            self.llmSettings = llm_settings.to_dict()

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "prompt": self.prompt,
            "llmSettings": self.llmSettings,
            "language": self.language,
            "indent": self.indent,
        }

    async def send(self):
        """
        Send the message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_message")
        id = await super().send()

        action_coros = [action.send(for_id=str(id)) for action in self.actions]
        element_coros = [element.send(for_id=str(id)) for element in self.elements]
        all_coros = action_coros + element_coros
        await asyncio.gather(*all_coros)

        return id


class ErrorMessage(MessageBase):
    """
    Send an error message to the UI
    If a project ID is configured, the message will be persisted in the cloud.

    Args:
        content (str): Text displayed above the upload button.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        indent (int, optional): If positive, the message will be nested in the UI.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        indent: int = 0,
    ):
        self.content = content
        self.author = author
        self.indent = indent

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "indent": self.indent,
            "isError": True,
        }

    async def send(self):
        """
        Send the error message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_error_message")
        return await super().send()


class AskMessageBase(MessageBase):
    def remove(self):
        removed = super().remove()
        if removed:
            self.emitter.clear_ask()


class AskUserMessage(AskMessageBase):
    """
    Ask for the user input before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the message will be uploaded to the cloud storage.

    Args:
        content (str): The content of the prompt.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        timeout: int = 60,
        raise_on_timeout: bool = False,
    ):
        self.content = content
        self.author = author
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timeout

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
        }

    async def send(self) -> Union[AskResponse, None]:
        """
        Sends the question to ask to the UI and waits for the reply.
        """
        trace_event("send_ask_user")

        if self.streaming:
            self.streaming = False

        msg_dict = await self._create()

        spec = AskSpec(type="text", timeout=self.timeout)

        res = await self.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout)

        return res


class AskFileMessage(AskMessageBase):
    """
    Ask the user to upload a file before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the file will be uploaded to the cloud storage.

    Args:
        content (str): Text displayed above the upload button.
        accept (Union[List[str], Dict[str, List[str]]]): List of mime type to accept like ["text/csv", "application/pdf"] or a dict like {"text/plain": [".txt", ".py"]}.
        max_size_mb (int, optional): Maximum size per file in MB. Maximum value is 100.
        max_files (int, optional): Maximum number of files to upload. Maximum value is 10.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    """

    def __init__(
        self,
        content: str,
        accept: Union[List[str], Dict[str, List[str]]],
        max_size_mb=2,
        max_files=1,
        author=config.ui.name,
        timeout=90,
        raise_on_timeout=False,
    ):
        self.content = content
        self.max_size_mb = max_size_mb
        self.max_files = max_files
        self.accept = accept
        self.author = author
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timeout

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
        }

    async def send(self) -> Union[List[AskFileResponse], None]:
        """
        Sends the message to request a file from the user to the UI and waits for the reply.
        """
        trace_event("send_ask_file")

        if self.streaming:
            self.streaming = False

        msg_dict = await self._create()

        spec = AskFileSpec(
            type="file",
            accept=self.accept,
            max_size_mb=self.max_size_mb,
            max_files=self.max_files,
            timeout=self.timeout,
        )

        res = await self.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout)

        if res:
            return [AskFileResponse(**r) for r in res]
        else:
            return None



def mock_openai():
    import os
    from aioresponses import aioresponses
    import responses

    # Mock the openai api key
    os.environ["OPENAI_API_KEY"] = "sk-FAKE-OPENAI-API-KEY"

    mocked_json_reply = {
        "id": "cmpl-uqkvlQyYK7bGYrRHQ0eXlWi7",
        "object": "text_completion",
        "created": 1589478378,
        "model": "text-davinci-003",
        "choices": [
            {
                "text": "\n\n```text\n3*3\n```",
                "index": 0,
                "logprobs": None,
                "finish_reason": "length",
            }
        ],
        "usage": {
            "prompt_tokens": 5,
            "completion_tokens": 7,
            "total_tokens": 12,
        },
    }

    # Mock the sync openai api

    responses.start()

    responses.add(
        responses.POST,
        "https://api.openai.com/v1/completions",
        json=mocked_json_reply,
    )

    # Mock the async openai api
    aioresponses_obj = aioresponses()

    aioresponses_obj.start()

    aioresponses_obj.add(
        url="https://api.openai.com/v1/completions",
        method="POST",
        payload=mocked_json_reply,
    )



from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json

from chainlit.emitter import get_emit_fn
from chainlit.telemetry import trace_event


@dataclass_json
@dataclass
class Action:
    # Name of the action, this should be used in the action_callback
    name: str
    # The value associated with the action. This is useful to differentiate between multiple actions with the same name.
    value: str
    # The label of the action. This is what the user will see. If not provided the name will be used.
    label: str = ""
    # The description of the action. This is what the user will see when they hover the action.
    description: str = ""
    # This should not be set manually, only used internally.
    forId: str = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.emit = get_emit_fn()
        if not self.emit:
            raise RuntimeError("Action should be instantiated in a Chainlit context")

    async def send(self, for_id: str):
        trace_event(f"send {self.__class__.__name__}")
        self.forId = for_id
        await self.emit("action", self.to_dict())

    async def remove(self):
        trace_event(f"remove {self.__class__.__name__}")
        await self.emit("remove_action", self.to_dict())



import os
import tarfile
import fnmatch
import tempfile
import requests
from chainlit.cli.utils import check_file
from chainlit.config import config
from chainlit.cli.auth import get_access_token


def get_gitignore_patterns():
    patterns = []
    if not os.path.exists(".gitignore"):
        return patterns

    with open(".gitignore", "r", encoding="utf-8") as gitignore_file:
        for line in gitignore_file:
            line = line.strip()
            if line and not line.startswith("#"):
                patterns.append(line)

    return patterns


def is_excluded(file_path, patterns):
    for pattern in patterns:
        if fnmatch.fnmatch(file_path, pattern):
            return True
    return False


def create_tar_gz_archive(archive_path, source_dir):
    gitignore_patterns = get_gitignore_patterns()
    with tarfile.open(archive_path, "w:gz") as tar_file:
        for root, dirs, files in os.walk(source_dir):
            for file in files:
                file_path = os.path.join(root, file)
                if not is_excluded(file_path, gitignore_patterns):
                    tar_file.add(
                        file_path, arcname=os.path.relpath(file_path, source_dir)
                    )


def upload_tar_gz_archive(access_token: str, archive_path: str):
    url = f"{config.chainlit_server}/api/upload/deployment"
    body = {"projectId": config.project.id}

    headers = {"Authorization": access_token}

    res = requests.post(url, json=body, headers=headers)

    if not res.ok:
        raise Exception(f"Failed to upload archive: {res.text}")

    json_res = res.json()

    upload_details = json_res["post"]
    permanent_url = json_res["permanentUrl"]

    files = {"file": open(archive_path, "rb")}

    upload_response = requests.post(
        upload_details["url"], data=upload_details["fields"], files=files
    )

    if not upload_response.ok:
        raise Exception(f"Failed to upload archive: {res.text}")

    url = f'{upload_details["url"]}/{upload_details["fields"]["key"]}'
    return permanent_url, url


def deploy(target: str):
    if not config.project.id:
        raise Exception(
            "Project id not set in config. A project id is mandatory to deploy a chainlit app."
        )

    check_file(target)

    access_token = get_access_token()

    with tempfile.TemporaryDirectory() as tempdir:
        archive_path = os.path.join(tempdir, "chainlit.tar.gz")
        source_dir = os.getcwd()
        create_tar_gz_archive(archive_path, source_dir)
        permanent_url, url = upload_tar_gz_archive(access_token, archive_path)



import uptrace
import hashlib
from socket import gethostname
from opentelemetry.sdk.resources import Attributes, Resource
from opentelemetry import trace as ot_trace
from opentelemetry.exporter.otlp.proto.grpc.exporter import logger as exporter_logger
from functools import wraps
from chainlit.config import config
from chainlit.version import __version__
import logging


# Patch uptrace.py to hash the hostname to avoid leaking it.
def _build_resource(
    resource: Resource,
    resource_attributes: Attributes,
    service_name: str,
    service_version: str,
    deployment_environment: str,
) -> Resource:
    if resource:
        return resource

    # If we are in production, use the URL as hostname
    if config.chainlit_prod_url:
        host_name = config.chainlit_prod_url
    # Hash the local hostname to avoid leaking it.
    else:
        host_name = gethostname()
        host_name = hashlib.sha256(host_name.encode("UTF-8")).hexdigest()

    attrs = {"host.name": host_name}

    if resource_attributes:
        attrs.update(resource_attributes)
    if service_name:
        attrs["service.name"] = service_name
    if service_version:
        attrs["service.version"] = service_version
    if deployment_environment:
        attrs["deployment.environment"] = deployment_environment

    return Resource.create(attrs)


uptrace.uptrace._build_resource = _build_resource

uptrace.configure_opentelemetry(
    dsn="https://YPa4AbDF853uCW6UWN2oYg@api.uptrace.dev/1778",
    service_name="chainlit",
    service_version="1.0.0",
    deployment_environment="production",
    logging_level=logging.CRITICAL,
)

exporter_logger.setLevel(logging.CRITICAL)

tracer = ot_trace.get_tracer("chainlit", __version__)


def trace_event(event_name):
    if config.project.enable_telemetry:
        with tracer.start_as_current_span(
            event_name, record_exception=False, set_status_on_exception=False
        ):
            return


def trace(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        event_name = func.__name__
        if config.project.enable_telemetry:
            with tracer.start_as_current_span(
                event_name, record_exception=False, set_status_on_exception=False
            ):
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper



from typing import Union, Dict
from chainlit.session import Session
from chainlit.types import AskSpec
from chainlit.client import BaseClient
from socketio.exceptions import TimeoutError
import inspect


class ChainlitEmitter:
    """
    Chainlit Emitter class. The Emitter is not directly exposed to the developer.
    Instead, the developer interacts with the Emitter through the methods and classes exposed in the __init__ file.
    """

    session: Session

    def __init__(self, session: Session) -> None:
        """Initialize with the user session."""
        self.session = session

    def _get_session_property(self, property_name: str, raise_error=True):
        """Helper method to get a property from the session."""
        if not hasattr(self, "session") or property_name not in self.session:
            if raise_error:
                raise ValueError(f"Session does not have property '{property_name}'")
            else:
                return None
        return self.session[property_name]

    @property
    def emit(self):
        """Get the 'emit' property from the session."""
        return self._get_session_property("emit")

    @property
    def ask_user(self):
        """Get the 'ask_user' property from the session."""
        return self._get_session_property("ask_user")

    @property
    def client(self) -> Union[BaseClient, None]:
        """Get the 'client' property from the session."""
        return self._get_session_property("client", raise_error=False)

    def send_message(self, msg_dict: Dict):
        """Send a message to the UI."""

        return self.emit("new_message", msg_dict)

    def update_message(self, msg_dict: Dict):
        """Update a message in the UI."""

        return self.emit("update_message", msg_dict)

    def delete_message(self, msg_dict):
        """Delete a message in the UI."""

        return self.emit("delete_message", msg_dict)

    def send_ask_timeout(self):
        """Send a prompt timeout message to the UI."""

        return self.emit("ask_timeout", {})

    def clear_ask(self):
        """Clear the prompt from the UI."""

        return self.emit("clear_ask", {})

    async def send_ask_user(
        self, msg_dict: Dict, spec: AskSpec, raise_on_timeout=False
    ):
        """Send a prompt to the UI and wait for a response."""

        try:
            # Send the prompt to the UI
            res = await self.ask_user(
                {"msg": msg_dict, "spec": spec.to_dict()}, spec.timeout
            )

            # End the task temporarily so that the User can answer the prompt
            await self.task_end()

            if self.client and res:
                # If cloud is enabled, store the response in the database/S3
                if spec.type == "text":
                    res_msg = {
                        "author": res["author"],
                        "authorIsUser": True,
                        "content": res["content"],
                    }
                    await self.client.create_message(res_msg)
                elif spec.type == "file":
                    # TODO: upload file to S3
                    pass

            await self.clear_ask()
            return res
        except TimeoutError as e:
            await self.send_ask_timeout()

            if raise_on_timeout:
                raise e
        finally:
            await self.task_start()

    def update_token_count(self, count: int):
        """Update the token count for the UI."""

        return self.emit("token_usage", count)

    def task_start(self):
        """
        Send a task start signal to the UI.
        """
        return self.emit("task_start", {})

    def task_end(self):
        """Send a task end signal to the UI."""
        return self.emit("task_end", {})

    def stream_start(self, msg_dict: Dict):
        """Send a stream start signal to the UI."""
        return self.emit(
            "stream_start",
            msg_dict,
        )

    def send_token(self, id: Union[str, int], token: str):
        """Send a message token to the UI."""
        return self.emit("stream_token", {"id": id, "token": token})


def get_emitter() -> Union[ChainlitEmitter, None]:
    """
    Get the Chainlit Emitter instance from the current call stack.
    This unusual approach is necessary because:
     - we need to get the right Emitter instance with the right websocket connection
     - to preserve a lean developer experience, we do not pass the Emitter instance to every function call

    What happens is that we set __chainlit_emitter__ in the local variables when we receive a websocket message.
    Then we can retrieve it from the call stack when we need it, even if the developer's code has no idea about it.
    """
    attr = "__chainlit_emitter__"
    candidates = [i[0].f_locals.get(attr) for i in inspect.stack()]
    emitter = None
    for candidate in candidates:
        if candidate:
            emitter = candidate
            break

    return emitter


def get_emit_fn():
    emitter = get_emitter()
    if emitter:
        return emitter.emit
    return None



from auth0.authentication.token_verifier import (
    TokenVerifier,
    AsymmetricSignatureVerifier,
)
from typing import Dict
from chainlit.logger import logger
import time
import requests
import sys
import time
import webbrowser
import os
import json


AUTH0_DOMAIN = "auth.chainlit.io"
AUTH0_CLIENT_ID = "ADo93BBXDn8Z35lEi8arCWiR7C0ncrjx"
ALGORITHMS = ["HS256"]


def get_credentials_path():
    """
    Returns the path to the credentials file
    """
    return os.path.join(os.path.expanduser("~"), ".chainlit", "credentials.json")


def store_credentials(credentials: Dict):
    """
    Stores the credentials in the credentials file

    :param credentials:
    """
    if not os.path.exists(os.path.dirname(get_credentials_path())):
        os.makedirs(os.path.dirname(get_credentials_path()))
    with open(get_credentials_path(), "w", encoding="utf-8") as f:
        f.write(json.dumps(credentials))


def is_logged_in():
    """
    Returns true if the user is logged in
    """
    if not os.path.exists(get_credentials_path()):
        return False
    with open(get_credentials_path(), "r", encoding="utf-8") as f:
        credentials = json.loads(f.read())
        if time.time() - credentials["created_at"] > credentials["expires_in"]:
            logger.info("Token expired.")
            return False
        return True


def get_access_token():
    """
    Returns the credentials from the credentials file
    """
    if not is_logged_in():
        login()
        return get_access_token()
    with open(get_credentials_path(), "r", encoding="utf-8") as f:
        credentials = json.loads(f.read())
        return credentials["access_token"]


def validate_token(token):
    """
    Verify the token and its precedence

    :param token:
    """
    jwks_url = "https://{}/.well-known/jwks.json".format(AUTH0_DOMAIN)
    issuer = "https://{}/".format(AUTH0_DOMAIN)
    sv = AsymmetricSignatureVerifier(jwks_url)
    tv = TokenVerifier(signature_verifier=sv, issuer=issuer, audience=AUTH0_CLIENT_ID)
    return tv.verify(token)


def logout():
    """
    Removes the user object from memory
    """
    if os.path.exists(get_credentials_path()):
        os.remove(get_credentials_path())
        logger.info("Logged out")
    else:
        logger.error("You are not logged in")


def login():
    """
    Runs the device authorization flow and stores the user object in memory
    """
    if is_logged_in():
        logger.info("You are already logged in")
        return

    device_code_payload = {
        "client_id": AUTH0_CLIENT_ID,
        "scope": "openid profile email",
        "audience": "chainlit-cloud",
    }
    device_code_response = requests.post(
        "https://{}/oauth/device/code".format(AUTH0_DOMAIN), data=device_code_payload
    )

    if device_code_response.status_code != 200:
        logger.error("Error generating the device code")
        logger.error(device_code_response.json())
        sys.exit(1)

    device_code_data = device_code_response.json()
    webbrowser.open(device_code_data["verification_uri_complete"])
    logger.info(f"Enter the following code: {device_code_data['user_code']}")

    token_payload = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code_data["device_code"],
        "client_id": AUTH0_CLIENT_ID,
    }

    authenticated = False
    while not authenticated:
        token_response = requests.post(
            "https://{}/oauth/token".format(AUTH0_DOMAIN), data=token_payload
        )

        token_data = token_response.json()
        if token_response.status_code == 200:
            user = validate_token(token_data["id_token"])
            token_data["created_at"] = time.time()
            store_credentials(token_data)
            logger.info(f"Logged in as {user['email']}")
            authenticated = True
        elif token_data["error"] not in ("authorization_pending", "slow_down"):
            logger.error(token_data["error_description"])
            sys.exit(1)
        else:
            time.sleep(device_code_data["interval"])



from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Dict, Union, Any
import uuid
import aiofiles
from io import BytesIO

from chainlit.emitter import get_emitter, BaseClient
from chainlit.telemetry import trace_event
from chainlit.types import ElementType, ElementDisplay, ElementSize

type_to_mime = {
    "image": "binary/octet-stream",
    "text": "text/plain",
    "pdf": "application/pdf",
}


@dataclass_json
@dataclass
class Element:
    # Name of the element, this will be used to reference the element in the UI.
    name: str
    # The type of the element. This will be used to determine how to display the element in the UI.
    type: ElementType
    # Controls how the image element should be displayed in the UI. Choices are “side” (default), “inline”, or “page”.
    display: ElementDisplay = "side"
    # The URL of the element if already hosted somehwere else.
    url: str = None
    # The local path of the element.
    path: str = None
    # The byte content of the element.
    content: bytes = None
    # The ID of the element. This is set automatically when the element is sent to the UI if cloud is enabled.
    id: int = None
    # The ID of the element if cloud is disabled.
    tempId: str = None
    # The ID of the message this element is associated with.
    forId: str = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.emitter = get_emitter()
        if not self.emitter:
            raise RuntimeError("Element should be instantiated in a Chainlit context")

        if not self.url and not self.path and not self.content:
            raise ValueError("Must provide url, path or content to instantiate element")

        self.tempId = uuid.uuid4().hex

    async def preprocess_content(self):
        pass

    async def load(self):
        if self.path:
            async with aiofiles.open(self.path, "rb") as f:
                self.content = await f.read()
                await self.preprocess_content()
        elif self.content:
            await self.preprocess_content()
        else:
            raise ValueError("Must provide path or content to load element")

    async def persist(self, client: BaseClient, for_id: str = None):
        if not self.url and self.content:
            self.url = await client.upload_element(
                content=self.content, mime=type_to_mime[self.type]
            )

        size = getattr(self, "size", None)
        language = getattr(self, "language", None)
        element = await client.create_element(
            name=self.name,
            url=self.url,
            type=self.type,
            display=self.display,
            size=size,
            language=language,
            for_id=for_id,
        )
        return element

    async def before_emit(self, element: Dict) -> Dict:
        return element

    async def send(self, for_id: str = None):
        element = None

        if not self.content and not self.url and self.path:
            await self.load()

        # Cloud is enabled, upload the element to S3
        if self.emitter.client and not self.id:
            element = await self.persist(self.emitter.client, for_id)
            self.id = element["id"]

        elif not self.url and not self.content:
            raise ValueError("Must provide url or content to send element")

        element = self.to_dict()
        if for_id:
            element["forId"] = for_id

        if self.emitter.emit and element:
            trace_event(f"send {self.__class__.__name__}")
            element = await self.before_emit(element)
            await self.emitter.emit("element", element)


@dataclass
class Image(Element):
    type: ElementType = "image"
    size: ElementSize = "medium"


@dataclass
class Avatar(Element):
    type: ElementType = "avatar"

    async def send(self):
        element = None

        if not self.content and not self.url and self.path:
            await self.load()

        if not self.url and not self.content:
            raise ValueError("Must provide url or content to send element")

        element = self.to_dict()

        if self.emitter.emit and element:
            trace_event(f"send {self.__class__.__name__}")
            element = await self.before_emit(element)
            await self.emitter.emit("element", element)


@dataclass
class Text(Element):
    """Useful to send a text (not a message) to the UI."""

    content: Union[str, bytes] = None
    type: ElementType = "text"
    language: str = None

    async def preprocess_content(self):
        if isinstance(self.content, str):
            self.content = self.content.encode("utf-8")

    async def before_emit(self, text_element):
        if "content" in text_element and isinstance(text_element["content"], bytes):
            text_element["content"] = text_element["content"].decode("utf-8")
        return text_element


@dataclass
class Pdf(Element):
    """Useful to send a pdf to the UI."""

    type: ElementType = "pdf"


@dataclass
class Pyplot(Element):
    """Useful to send a pyplot to the UI."""

    # We reuse the frontend image element to display the chart
    type: ElementType = "image"
    # The type is set to Any because the figure is not serializable
    # and its actual type is checked in __post_init__.
    figure: Any = None

    def __post_init__(self) -> None:
        from matplotlib.figure import Figure

        if not isinstance(self.figure, Figure):
            raise TypeError("figure must be a matplotlib.figure.Figure")

        options = {
            "dpi": 200,
            "bbox_inches": "tight",
            "backend": "Agg",
            "format": "png",
        }
        image = BytesIO()
        self.figure.savefig(image, **options)
        self.content = image.getvalue()

        super().__post_init__()

    async def before_emit(self, element: Dict) -> Dict:
        # Prevent the figure from being serialized
        del element["figure"]
        return element



from dotenv import load_dotenv
from typing import Callable, Any
import inspect
import os
import asyncio

from chainlit.lc import LANGCHAIN_INSTALLED
from chainlit.config import config
from chainlit.telemetry import trace
from chainlit.version import __version__
from chainlit.logger import logger
from chainlit.emitter import ChainlitEmitter
from chainlit.types import LLMSettings
from chainlit.message import ErrorMessage
from chainlit.action import Action
from chainlit.element import Image, Text, Pdf, Avatar, Pyplot
from chainlit.message import Message, ErrorMessage, AskUserMessage, AskFileMessage
from chainlit.user_session import user_session
from chainlit.sync import run_sync, make_async
from chainlit.cache import cache

if LANGCHAIN_INSTALLED:
    from chainlit.lc.callbacks import (
        ChainlitCallbackHandler,
        AsyncChainlitCallbackHandler,
    )


env_found = load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))

if env_found:
    logger.info("Loaded .env file")


def wrap_user_function(user_function: Callable, with_task=False) -> Callable:
    """
    Wraps a user-defined function to accept arguments as a dictionary.

    Args:
        user_function (Callable): The user-defined function to wrap.

    Returns:
        Callable: The wrapped function.
    """

    async def wrapper(*args, __chainlit_emitter__: ChainlitEmitter):
        # Get the parameter names of the user-defined function
        user_function_params = list(inspect.signature(user_function).parameters.keys())

        # Create a dictionary of parameter names and their corresponding values from *args
        params_values = {
            param_name: arg for param_name, arg in zip(user_function_params, args)
        }

        if with_task:
            await __chainlit_emitter__.task_start()

        try:
            # Call the user-defined function with the arguments
            if inspect.iscoroutinefunction(user_function):
                return await user_function(**params_values)
            else:
                return user_function(**params_values)
        except InterruptedError:
            pass
        except Exception as e:
            logger.exception(e)
            await ErrorMessage(content=str(e), author="Error").send()
        finally:
            if with_task:
                await __chainlit_emitter__.task_end()

    return wrapper


@trace
def langchain_factory(use_async: bool) -> Callable:
    """
    Plug and play decorator for the LangChain library.
    The decorated function should instantiate a new LangChain instance (Chain, Agent...).
    One instance per user session is created and cached.
    The per user instance is called every time a new message is received.

    Args:
        use_async bool: Whether to call the the agent asynchronously or not. Defaults to False.

    Returns:
        Callable[[], Any]: The decorated factory function.
    """

    # Check if the factory is called with the correct parameter
    if type(use_async) != bool:
        error_message = "langchain_factory use_async parameter is required"
        raise ValueError(error_message)

    def decorator(func: Callable) -> Callable:
        config.code.lc_factory = wrap_user_function(func, with_task=True)
        return func

    config.code.lc_agent_is_async = use_async

    return decorator


@trace
def langchain_postprocess(func: Callable[[Any], str]) -> Callable:
    """
    Useful to post process the response a LangChain object instantiated with @langchain_factory.
    The decorated function takes the raw output of the LangChain object as input.
    The response will NOT be automatically sent to the UI, you need to send a Message.

    Args:
        func (Callable[[Any], str]): The post-processing function to apply after generating a response. Takes the response as parameter.

    Returns:
        Callable[[Any], str]: The decorated post-processing function.
    """

    config.code.lc_postprocess = wrap_user_function(func)
    return func


@trace
def on_message(func: Callable) -> Callable:
    """
    Framework agnostic decorator to react to messages coming from the UI.
    The decorated function is called every time a new message is received.

    Args:
        func (Callable[[str], Any]): The function to be called when a new message is received. Takes the input message.

    Returns:
        Callable[[str], Any]: The decorated on_message function.
    """

    config.code.on_message = wrap_user_function(func)
    return func


@trace
def langchain_run(func: Callable[[Any, str], str]) -> Callable:
    """
    Useful to override the default behavior of the LangChain object instantiated with @langchain_factory.
    Use when your agent run method has custom parameters.
    Takes the LangChain agent and the user input as parameters.
    The response will NOT be automatically sent to the UI, you need to send a Message.
    Args:
        func (Callable[[Any, str], str]): The function to be called when a new message is received. Takes the agent and user input as parameters and returns the output string.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """
    config.code.lc_run = wrap_user_function(func)
    return func


@trace
def langchain_rename(func: Callable[[str], str]) -> Callable[[str], str]:
    """
    Useful to rename the LangChain tools/chains used in the agent and display more friendly author names in the UI.
    Args:
        func (Callable[[str], str]): The function to be called to rename a tool/chain. Takes the original tool/chain name as parameter.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """

    config.code.lc_rename = wrap_user_function(func)
    return func


@trace
def on_chat_start(func: Callable) -> Callable:
    """
    Hook to react to the user websocket connection event.

    Args:
        func (Callable[], Any]): The connection hook to execute.

    Returns:
        Callable[], Any]: The decorated hook.
    """

    config.code.on_chat_start = wrap_user_function(func, with_task=True)
    return func


@trace
def on_stop(func: Callable) -> Callable:
    """
    Hook to react to the user stopping a conversation.

    Args:
        func (Callable[[], Any]): The stop hook to execute.

    Returns:
        Callable[[], Any]: The decorated stop hook.
    """

    config.code.on_stop = wrap_user_function(func)
    return func


def action_callback(name: str) -> Callable:
    """
    Callback to call when an action is clicked in the UI.

    Args:
        func (Callable[[Action], Any]): The action callback to exexute. First parameter is the action.
    """

    def decorator(func: Callable[[Action], Any]):
        config.code.action_callbacks[name] = wrap_user_function(func, with_task=True)
        return func

    return decorator


def sleep(duration: int):
    """
    Sleep for a given duration.
    Args:
        duration (int): The duration in seconds.
    """
    return asyncio.sleep(duration)


__all__ = [
    "user_session",
    "LLMSettings",
    "Action",
    "Pdf",
    "Image",
    "Text",
    "Avatar",
    "Pyplot",
    "Message",
    "ErrorMessage",
    "AskUserMessage",
    "AskFileMessage",
    "langchain_factory",
    "langchain_postprocess",
    "langchain_run",
    "langchain_rename",
    "on_chat_start",
    "on_stop",
    "action_callback",
    "sleep",
    "ChainlitCallbackHandler",
    "AsyncChainlitCallbackHandler",
    "run_sync",
    "make_async",
    "cache",
]



import os
import click


def check_file(target: str):
    # Define accepted file extensions for Chainlit
    ACCEPTED_FILE_EXTENSIONS = ("py", "py3")

    _, extension = os.path.splitext(target)

    # Check file extension
    if extension[1:] not in ACCEPTED_FILE_EXTENSIONS:
        if extension[1:] == "":
            raise click.BadArgumentUsage(
                "Chainlit requires raw Python (.py) files, but the provided file has no extension."
            )
        else:
            raise click.BadArgumentUsage(
                f"Chainlit requires raw Python (.py) files, not {extension}."
            )

    if not os.path.exists(target):
        raise click.BadParameter(f"File does not exist: {target}")



try:
    import langchain

    if langchain.__version__ < "0.0.198":
        raise ValueError(
            "LangChain version is too old, expected >= 0.0.198. Run `pip install langchain --upgrade`"
        )

    LANGCHAIN_INSTALLED = True
except ImportError:
    LANGCHAIN_INSTALLED = False



from typing import Any
from chainlit.lc.callbacks import ChainlitCallbackHandler, AsyncChainlitCallbackHandler
from chainlit.sync import make_async


async def run_langchain_agent(agent: Any, input_str: str, use_async: bool):
    if hasattr(agent, "input_keys"):
        input_key = agent.input_keys[0]
        if use_async:
            raw_res = await agent.acall(
                {input_key: input_str}, callbacks=[AsyncChainlitCallbackHandler()]
            )
        else:
            raw_res = await make_async(agent.__call__)(
                {input_key: input_str}, callbacks=[ChainlitCallbackHandler()]
            )
    else:
        if use_async:
            raw_res = await agent.acall(
                input_str, callbacks=[AsyncChainlitCallbackHandler()]
            )
        else:
            raw_res = await make_async(agent.__call__)(
                input_str, callbacks=[ChainlitCallbackHandler()]
            )

    if hasattr(agent, "output_keys"):
        output_key = agent.output_keys[0]
    else:
        output_key = None

    return raw_res, output_key



from typing import List, TypedDict, Optional, Literal, Dict, Union
from pydantic import BaseModel
from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json

ElementType = Literal["image", "avatar", "text", "pdf"]
ElementDisplay = Literal["inline", "side", "page"]
ElementSize = Literal["small", "medium", "large"]


@dataclass_json
@dataclass
class AskSpec:
    """Specification for asking the user."""

    timeout: int
    type: Literal["text", "file"]


@dataclass_json
@dataclass
class AskFileSpec(AskSpec):
    """Specification for asking the user for a file."""

    accept: Union[List[str], Dict[str, List[str]]]
    max_files: int
    max_size_mb: int


class AskResponse(TypedDict):
    content: str
    author: str


@dataclass
class AskFileResponse:
    name: str
    path: str
    size: int
    type: str
    content: bytes


@dataclass_json
@dataclass
class LLMSettings:
    model_name: str = "text-davinci-003"
    stop: Optional[List[str]] = None
    temperature: float = 0
    max_tokens: int = 256
    top_p: int = 1
    frequency_penalty: int = 0
    presence_penalty: int = 0

    def to_settings_dict(self):
        return {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
        }


class CompletionRequest(BaseModel):
    prompt: str
    userEnv: Dict[str, str]
    settings: LLMSettings



from typing import Dict, Any, Optional
from abc import ABC, abstractmethod
import uuid

import asyncio
import aiohttp
from python_graphql_client import GraphqlClient

from chainlit.types import ElementType, ElementSize
from chainlit.logger import logger
from chainlit.config import config


class BaseClient(ABC):
    project_id: str
    session_id: str

    @abstractmethod
    async def is_project_member(self, access_token: str) -> bool:
        pass

    @abstractmethod
    async def create_conversation(self, session_id: str) -> int:
        pass

    @abstractmethod
    async def create_message(self, variables: Dict[str, Any]) -> int:
        pass

    @abstractmethod
    async def update_message(self, message_id: int, variables: Dict[str, Any]) -> bool:
        pass

    @abstractmethod
    async def delete_message(self, message_id: int) -> bool:
        pass

    @abstractmethod
    async def upload_element(self, content: bytes, mime: str) -> str:
        pass

    @abstractmethod
    async def create_element(
        self,
        type: ElementType,
        url: str,
        name: str,
        display: str,
        size: ElementSize = None,
        language: str = None,
        for_id: str = None,
    ) -> Dict[str, Any]:
        pass


conversation_lock = asyncio.Lock()


class CloudClient(BaseClient):
    conversation_id: Optional[str] = None

    def __init__(self, project_id: str, session_id: str, access_token: str):
        self.project_id = project_id
        self.session_id = session_id
        self.headers = {
            "Authorization": access_token,
            "content-type": "application/json",
        }
        graphql_endpoint = f"{config.chainlit_server}/api/graphql"
        self.graphql_client = GraphqlClient(
            endpoint=graphql_endpoint, headers=self.headers
        )

    def query(self, query: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL query.

        :param query: The GraphQL query string.
        :param variables: A dictionary of variables for the query.
        :return: The response data as a dictionary.
        """
        return self.graphql_client.execute_async(query=query, variables=variables)

    def check_for_errors(self, response: Dict[str, Any]):
        if "errors" in response:
            logger.error(response["errors"])
            return True
        return False

    def mutation(self, mutation: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL mutation.

        :param mutation: The GraphQL mutation string.
        :param variables: A dictionary of variables for the mutation.
        :return: The response data as a dictionary.
        """
        return self.graphql_client.execute_async(query=mutation, variables=variables)

    async def is_project_member(self) -> bool:
        data = {"projectId": self.project_id}
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{config.chainlit_server}/api/role",
                json=data,
                headers=self.headers,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    logger.error(f"Failed to get user role. {r.status}: {reason}")
                    return False
                json = await r.json()
                role = json.get("role", "ANONYMOUS")
                return role != "ANONYMOUS"

    async def create_conversation(self, session_id: str) -> int:
        # If we run multiple send concurrently, we need to make sure we don't create multiple conversations.
        async with conversation_lock:
            if self.conversation_id:
                return self.conversation_id

            mutation = """
            mutation ($projectId: String!, $sessionId: String!) {
                createConversation(projectId: $projectId, sessionId: $sessionId) {
                    id
                }
            }
            """
            variables = {"projectId": self.project_id, "sessionId": session_id}
            res = await self.mutation(mutation, variables)

            if self.check_for_errors(res):
                logger.warning("Could not create conversation.")
                return None

            return int(res["data"]["createConversation"]["id"])

    async def get_conversation_id(self):
        self.conversation_id = await self.create_conversation(self.session_id)

        return self.conversation_id

    async def create_message(self, variables: Dict[str, Any]) -> int:
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the message.")
            return None

        variables["conversationId"] = c_id

        mutation = """
        mutation ($conversationId: ID!, $author: String!, $content: String!, $language: String, $prompt: String, $llmSettings: Json, $isError: Boolean, $indent: Int, $authorIsUser: Boolean, $waitForAnswer: Boolean, $createdAt: Float) {
            createMessage(conversationId: $conversationId, author: $author, content: $content, language: $language, prompt: $prompt, llmSettings: $llmSettings, isError: $isError, indent: $indent, authorIsUser: $authorIsUser, waitForAnswer: $waitForAnswer, createdAt: $createdAt) {
                id
            }
        }
        """
        res = await self.mutation(mutation, variables)
        if self.check_for_errors(res):
            logger.warning("Could not create message.")
            return None

        return int(res["data"]["createMessage"]["id"])

    async def update_message(self, message_id: int, variables: Dict[str, Any]) -> bool:
        mutation = """
        mutation ($messageId: ID!, $author: String!, $content: String!, $language: String, $prompt: String, $llmSettings: Json) {
            updateMessage(messageId: $messageId, author: $author, content: $content, language: $language, prompt: $prompt, llmSettings: $llmSettings) {
                id
            }
        }
        """
        variables["messageId"] = message_id
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not update message.")
            return False

        return True

    async def delete_message(self, message_id: int) -> bool:
        mutation = """
        mutation ($messageId: ID!) {
            deleteMessage(messageId: $messageId) {
                id
            }
        }
        """
        res = await self.mutation(mutation, {"messageId": message_id})

        if self.check_for_errors(res):
            logger.warning("Could not delete message.")
            return False

        return True

    async def create_element(
        self,
        type: ElementType,
        url: str,
        name: str,
        display: str,
        size: ElementSize = None,
        language: str = None,
        for_id: str = None,
    ) -> Dict[str, Any]:
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the element.")
            return None

        mutation = """
        mutation ($conversationId: ID!, $type: String!, $url: String!, $name: String!, $display: String!, $size: String, $language: String, $forId: String) {
            createElement(conversationId: $conversationId, type: $type, url: $url, name: $name, display: $display, size: $size, language: $language, forId: $forId) {
                id,
                type,
                url,
                name,
                display,
                size,
                language,
                forId
            }
        }
        """
        variables = {
            "conversationId": c_id,
            "type": type,
            "url": url,
            "name": name,
            "display": display,
            "size": size,
            "language": language,
            "forId": for_id,
        }
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not persist element.")
            return None

        return res["data"]["createElement"]

    async def upload_element(self, content: bytes, mime: str) -> str:
        id = f"{uuid.uuid4()}"
        body = {"projectId": self.project_id, "fileName": id, "contentType": mime}

        path = f"/api/upload/file"

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{config.chainlit_server}{path}",
                json=body,
                headers=self.headers,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    logger.error(f"Failed to upload file: {reason}")
                    return ""
                json_res = await r.json()

        upload_details = json_res["post"]
        permanent_url = json_res["permanentUrl"]

        form_data = aiohttp.FormData()

        # Add fields to the form_data
        for field_name, field_value in upload_details["fields"].items():
            form_data.add_field(field_name, field_value)

        # Add file to the form_data
        form_data.add_field("file", content, content_type="multipart/form-data")
        async with aiohttp.ClientSession() as session:
            async with session.post(
                upload_details["url"],
                data=form_data,
            ) as upload_response:
                if not upload_response.ok:
                    reason = await upload_response.text()
                    logger.error(f"Failed to upload file: {reason}")
                    return ""

                url = f'{upload_details["url"]}/{upload_details["fields"]["key"]}'
                return permanent_url



from typing import Dict, TypedDict, Optional, Callable, Any, Union
from chainlit.client import BaseClient
from chainlit.types import AskResponse


class Session(TypedDict):
    """Internal session object."""

    # The ID of the session
    id: str
    # Function to ask the user a question
    ask_user: Callable[[Any, Optional[int]], Union[AskResponse, None]]
    # Function to emit a message to the user
    emit: Callable[[str, Any], None]
    # User specific environment variables. Empty if no user environment variables are required.
    user_env: Dict[str, str]
    # Optional langchain agent
    agent: Any
    # If the session is currently running a sync task
    running_sync: bool
    # Whether the current task should be stopped
    should_stop: bool
    # Optional client to persist messages and files
    client: Optional[BaseClient]


sessions: Dict[str, Session] = {}



import logging

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)

logging.getLogger("socketio").setLevel(logging.ERROR)
logging.getLogger("engineio").setLevel(logging.ERROR)
logging.getLogger("geventwebsocket.handler").setLevel(logging.ERROR)
logging.getLogger("numexpr").setLevel(logging.ERROR)

logger = logging.getLogger("chainlit")



import mimetypes

mimetypes.add_type("application/javascript", ".js")
mimetypes.add_type("text/css", ".css")

import os
import json
import webbrowser

from contextlib import asynccontextmanager
from watchfiles import awatch

from fastapi import FastAPI
from fastapi.responses import (
    HTMLResponse,
    JSONResponse,
    FileResponse,
    PlainTextResponse,
)
from fastapi_socketio import SocketManager
from starlette.middleware.cors import CORSMiddleware
import asyncio

from chainlit.config import config, load_module, reload_config, DEFAULT_HOST
from chainlit.session import Session, sessions
from chainlit.user_session import user_sessions
from chainlit.client import CloudClient
from chainlit.emitter import ChainlitEmitter
from chainlit.markdown import get_markdown_str
from chainlit.action import Action
from chainlit.message import Message, ErrorMessage
from chainlit.telemetry import trace_event
from chainlit.logger import logger
from chainlit.types import CompletionRequest


@asynccontextmanager
async def lifespan(app: FastAPI):
    host = config.run.host
    port = config.run.port

    if not config.run.headless:
        if host == DEFAULT_HOST:
            url = f"http://localhost:{port}"
        else:
            url = f"http://{host}:{port}"

        logger.info(f"Your app is available at {url}")

        # Add a delay before opening the browser
        await asyncio.sleep(1)
        webbrowser.open(url)

    watch_task = None
    stop_event = asyncio.Event()

    if config.run.watch:

        async def watch_files_for_changes():
            extensions = [".py"]
            files = ["chainlit.md", "config.toml"]
            async for changes in awatch(config.root, stop_event=stop_event):
                for change_type, file_path in changes:
                    file_name = os.path.basename(file_path)
                    file_ext = os.path.splitext(file_name)[1]

                    if file_ext.lower() in extensions or file_name.lower() in files:
                        logger.info(
                            f"File {change_type.name}: {file_name}. Reloading app..."
                        )

                        try:
                            reload_config()
                        except Exception as e:
                            logger.error(f"Error reloading config: {e}")
                            break

                        # Reload the module if the module name is specified in the config
                        if config.run.module_name:
                            try:
                                load_module(config.run.module_name)
                            except Exception as e:
                                logger.error(f"Error reloading module: {e}")
                                break

                        await socket.emit("reload", {})

                        break

        watch_task = asyncio.create_task(watch_files_for_changes())

    try:
        yield
    finally:
        if watch_task:
            try:
                stop_event.set()
                watch_task.cancel()
                await watch_task
            except asyncio.exceptions.CancelledError:
                pass


root_dir = os.path.dirname(os.path.abspath(__file__))
build_dir = os.path.join(root_dir, "frontend/dist")

app = FastAPI(lifespan=lifespan)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Define max HTTP data size to 100 MB
max_http_data_size = 100 * 1024 * 1024

socket = SocketManager(
    app,
    cors_allowed_origins=[],
    async_mode="asgi",
    max_http_buffer_size=max_http_data_size,
)

"""
-------------------------------------------------------------------------------
                              HTTP HANDLERS
-------------------------------------------------------------------------------
"""


def get_html_template():
    PLACEHOLDER = "<!-- TAG INJECTION PLACEHOLDER -->"

    default_url = "https://github.com/Chainlit/chainlit"
    url = config.ui.github or default_url

    tags = f"""<title>{config.ui.name}</title>
    <meta name="description" content="{config.ui.description}">
    <meta property="og:type" content="website">
    <meta property="og:title" content="{config.ui.name}">
    <meta property="og:description" content="{config.ui.description}">
    <meta property="og:image" content="https://chainlit-cloud.s3.eu-west-3.amazonaws.com/logo/chainlit_banner.png">
    <meta property="og:url" content="{url}">"""

    index_html_file_path = os.path.join(build_dir, "index.html")

    with open(index_html_file_path, "r", encoding="utf-8") as f:
        content = f.read()
        content = content.replace(PLACEHOLDER, tags)
        return content


html_template = get_html_template()


@app.post("/completion")
async def completion(completion: CompletionRequest):
    """Handle a completion request from the prompt playground."""

    import openai

    trace_event("completion")

    api_key = completion.userEnv.get("OPENAI_API_KEY", os.environ.get("OPENAI_API_KEY"))

    model_name = completion.settings.model_name
    stop = completion.settings.stop
    # OpenAI doesn't support an empty stop array, clear it
    if isinstance(stop, list) and len(stop) == 0:
        stop = None

    if model_name in ["gpt-3.5-turbo", "gpt-4"]:
        response = await openai.ChatCompletion.acreate(
            api_key=api_key,
            model=model_name,
            messages=[{"role": "user", "content": completion.prompt}],
            stop=stop,
            **completion.settings.to_settings_dict(),
        )
        return PlainTextResponse(content=response["choices"][0]["message"]["content"])
    else:
        response = await openai.Completion.acreate(
            api_key=api_key,
            model=model_name,
            prompt=completion.prompt,
            stop=stop,
            **completion.settings.to_settings_dict(),
        )
        return PlainTextResponse(content=response["choices"][0]["text"])


@app.get("/project/settings")
async def project_settings():
    """Return project settings. This is called by the UI before the establishing the websocket connection."""
    return JSONResponse(
        content={
            "chainlitServer": config.chainlit_server,
            "prod": bool(config.chainlit_prod_url),
            "ui": config.ui.to_dict(),
            "project": config.project.to_dict(),
            "markdown": get_markdown_str(config.root),
        }
    )


@app.get("/{path:path}")
async def serve(path: str):
    """Serve the UI."""
    path_to_file = os.path.join(build_dir, path)
    if path != "" and os.path.exists(path_to_file):
        return FileResponse(path_to_file)
    else:
        return HTMLResponse(content=html_template, status_code=200)


"""
-------------------------------------------------------------------------------
                              WEBSOCKET HANDLERS
-------------------------------------------------------------------------------
"""


def need_session(id: str):
    """Return the session with the given id."""

    session = sessions.get(id)
    if not session:
        raise ValueError("Session not found")
    return session


@socket.on("connect")
async def connect(sid, environ):
    user_env = environ.get("HTTP_USER_ENV")
    authorization = environ.get("HTTP_AUTHORIZATION")
    cloud_client = None

    # Check authorization
    if not config.project.public and not authorization:
        # Refuse connection if the app is private and no access token is provided
        trace_event("no_access_token")
        logger.error("Connection refused: No access token provided")
        return False
    elif authorization and config.project.id:
        # Create the cloud client
        cloud_client = CloudClient(
            project_id=config.project.id,
            session_id=sid,
            access_token=authorization,
        )
        is_project_member = await cloud_client.is_project_member()
        if not is_project_member:
            logger.error("Connection refused: You are not a member of this project")
            return False

    # Check user env
    if config.project.user_env:
        # Check if requested user environment variables are provided
        if user_env:
            user_env = json.loads(user_env)
            for key in config.project.user_env:
                if key not in user_env:
                    trace_event("missing_user_env")
                    logger.error(
                        "Connection refused: Missing user environment variable: " + key
                    )
                    return False
        else:
            logger.error("Connection refused: Missing user environment variables")
            return False

    # Create the session

    # Function to send a message to this particular session
    def emit_fn(event, data):
        if sid in sessions:
            if sessions[sid]["should_stop"]:
                sessions[sid]["should_stop"] = False
                raise InterruptedError("Task stopped by user")
        return socket.emit(event, data, to=sid)

    # Function to ask the user a question
    def ask_user_fn(data, timeout):
        if sessions[sid]["should_stop"]:
            sessions[sid]["should_stop"] = False
            raise InterruptedError("Task stopped by user")
        return socket.call("ask", data, timeout=timeout, to=sid)

    session = {
        "id": sid,
        "emit": emit_fn,
        "ask_user": ask_user_fn,
        "client": cloud_client,
        "user_env": user_env,
        "running_sync": False,
        "should_stop": False,
    }  # type: Session

    sessions[sid] = session

    trace_event("connection_successful")
    return True


@socket.on("connection_successful")
async def connection_successful(sid):
    session = need_session(sid)
    __chainlit_emitter__ = ChainlitEmitter(session)
    if config.code.lc_factory:
        """Instantiate the langchain agent and store it in the session."""
        agent = await config.code.lc_factory(__chainlit_emitter__=__chainlit_emitter__)
        session["agent"] = agent

    if config.code.on_chat_start:
        """Call the on_chat_start function provided by the developer."""
        await config.code.on_chat_start(__chainlit_emitter__=__chainlit_emitter__)


@socket.on("disconnect")
async def disconnect(sid):
    if sid in sessions:
        # Clean up the session
        sessions.pop(sid)

    if sid in user_sessions:
        # Clean up the user session
        user_sessions.pop(sid)


@socket.on("stop")
async def stop(sid):
    if sid in sessions:
        trace_event("stop_task")
        session = sessions[sid]

        __chainlit_emitter__ = ChainlitEmitter(session)

        await Message(author="System", content="Task stopped by the user.").send()

        session["should_stop"] = True

        if config.code.on_stop:
            await config.code.on_stop()


async def process_message(session: Session, author: str, input_str: str):
    """Process a message from the user."""

    try:
        __chainlit_emitter__ = ChainlitEmitter(session)
        await __chainlit_emitter__.task_start()

        if session["client"]:
            # If cloud is enabled, persist the message
            await session["client"].create_message(
                {
                    "author": author,
                    "content": input_str,
                    "authorIsUser": True,
                }
            )

        langchain_agent = session.get("agent")
        if langchain_agent:
            from chainlit.lc.agent import run_langchain_agent

            # If a langchain agent is available, run it
            if config.code.lc_run:
                # If the developer provided a custom run function, use it
                await config.code.lc_run(
                    langchain_agent,
                    input_str,
                    __chainlit_emitter__=__chainlit_emitter__,
                )
                return
            else:
                # Otherwise, use the default run function
                raw_res, output_key = await run_langchain_agent(
                    langchain_agent, input_str, use_async=config.code.lc_agent_is_async
                )

                if config.code.lc_postprocess:
                    # If the developer provided a custom postprocess function, use it
                    await config.code.lc_postprocess(
                        raw_res, __chainlit_emitter__=__chainlit_emitter__
                    )
                    return
                elif output_key is not None:
                    # Use the output key if provided
                    res = raw_res[output_key]
                else:
                    # Otherwise, use the raw response
                    res = raw_res
            # Finally, send the response to the user
            await Message(author=config.ui.name, content=res).send()

        elif config.code.on_message:
            # If no langchain agent is available, call the on_message function provided by the developer
            await config.code.on_message(
                input_str, __chainlit_emitter__=__chainlit_emitter__
            )
    except InterruptedError:
        pass
    except Exception as e:
        logger.exception(e)
        await ErrorMessage(author="Error", content=str(e)).send()
    finally:
        await __chainlit_emitter__.task_end()


@socket.on("ui_message")
async def message(sid, data):
    """Handle a message sent by the User."""
    session = need_session(sid)
    session["should_stop"] = False

    input_str = data["content"].strip()
    author = data["author"]

    await process_message(session, author, input_str)


async def process_action(session: Session, action: Action):
    __chainlit_emitter__ = ChainlitEmitter(session)
    callback = config.code.action_callbacks.get(action.name)
    if callback:
        await callback(action, __chainlit_emitter__=__chainlit_emitter__)
    else:
        logger.warning("No callback found for action %s", action.name)


@socket.on("action_call")
async def call_action(sid, action):
    """Handle an action call from the UI."""
    session = need_session(sid)

    __chainlit_emitter__ = ChainlitEmitter(session)
    action = Action(**action)

    await process_action(session, action)



import click
import os
import sys
import uvicorn
import asyncio
import nest_asyncio

nest_asyncio.apply()

from chainlit.config import (
    config,
    init_config,
    load_module,
    PACKAGE_ROOT,
    DEFAULT_HOST,
    DEFAULT_PORT,
)
from chainlit.markdown import init_markdown
from chainlit.cli.auth import login, logout
from chainlit.cli.deploy import deploy
from chainlit.cli.utils import check_file
from chainlit.telemetry import trace_event
from chainlit.cache import init_lc_cache
from chainlit.logger import logger
from chainlit.server import app


# Create the main command group for Chainlit CLI
@click.group(context_settings={"auto_envvar_prefix": "CHAINLIT"})
@click.version_option(prog_name="Chainlit")
def cli():
    return


# Define the function to run Chainlit with provided options
def run_chainlit(target: str):
    host = os.environ.get("CHAINLIT_HOST", DEFAULT_HOST)
    port = int(os.environ.get("CHAINLIT_PORT", DEFAULT_PORT))
    config.run.host = host
    config.run.port = port

    check_file(target)
    # Load the module provided by the user
    config.run.module_name = target
    load_module(config.run.module_name)

    # Create the chainlit.md file if it doesn't exist
    init_markdown(config.root)

    # Initialize the LangChain cache if installed and enabled
    init_lc_cache()

    log_level = "debug" if config.run.debug else "error"

    # Start the server
    async def start():
        config = uvicorn.Config(app, host=host, port=port, log_level=log_level)
        server = uvicorn.Server(config)
        await server.serve()

    # Run the asyncio event loop instead of uvloop to enable re entrance
    asyncio.run(start())
    # uvicorn.run(app, host=host, port=port)


# Define the "run" command for Chainlit CLI
@cli.command("run")
@click.argument("target", required=True, envvar="RUN_TARGET")
@click.option("-w", "--watch", default=False, is_flag=True, envvar="WATCH")
@click.option("-h", "--headless", default=False, is_flag=True, envvar="HEADLESS")
@click.option("-d", "--debug", default=False, is_flag=True, envvar="DEBUG")
@click.option("-c", "--ci", default=False, is_flag=True, envvar="CI")
@click.option("--no-cache", default=False, is_flag=True, envvar="NO_CACHE")
@click.option("--host")
@click.option("--port")
def chainlit_run(target, watch, headless, debug, ci, no_cache, host, port):
    if host:
        os.environ["CHAINLIT_HOST"] = host
    if port:
        os.environ["CHAINLIT_PORT"] = port
    if ci:
        logger.info("Running in CI mode")
        config.project.enable_telemetry = False
        no_cache = True
        from chainlit.cli.mock import mock_openai

        mock_openai()

    else:
        trace_event("chainlit run")

    config.run.headless = headless
    config.run.debug = debug
    config.run.no_cache = no_cache
    config.run.ci = ci
    config.run.watch = watch

    run_chainlit(target)


@cli.command("deploy")
@click.argument("target", required=True, envvar="CHAINLIT_RUN_TARGET")
@click.argument("args", nargs=-1)
def chainlit_deploy(target, args=None, **kwargs):
    trace_event("chainlit deploy")
    raise NotImplementedError("Deploy is not yet implemented")
    deploy(target)


@cli.command("hello")
@click.argument("args", nargs=-1)
def chainlit_hello(args=None, **kwargs):
    trace_event("chainlit hello")
    hello_path = os.path.join(PACKAGE_ROOT, "hello.py")
    run_chainlit(hello_path)


@cli.command("login")
@click.argument("args", nargs=-1)
def chainlit_login(args=None, **kwargs):
    trace_event("chainlit login")
    login()
    sys.exit(0)


@cli.command("logout")
@click.argument("args", nargs=-1)
def chainlit_logout(args=None, **kwargs):
    trace_event("chainlit logout")
    logout()
    sys.exit(0)


@cli.command("init")
@click.argument("args", nargs=-1)
def chainlit_init(args=None, **kwargs):
    trace_event("chainlit init")
    init_config(log=True)



from typing import Any, Dict, List, Optional, Union
from langchain.callbacks.base import BaseCallbackHandler, AsyncCallbackHandler
from langchain.schema import (
    AgentAction,
    AgentFinish,
    BaseMessage,
    LLMResult,
)
from chainlit.emitter import get_emitter, ChainlitEmitter
from chainlit.message import Message, ErrorMessage
from chainlit.config import config
from chainlit.types import LLMSettings
from chainlit.sync import run_sync

IGNORE_LIST = ["AgentExecutor"]


def get_llm_settings(invocation_params: Union[Dict, None]):
    if invocation_params is None:
        return None
    elif invocation_params["_type"] == "openai":
        return LLMSettings(
            model_name=invocation_params["model_name"],
            stop=invocation_params["stop"],
            temperature=invocation_params["temperature"],
            max_tokens=invocation_params["max_tokens"],
            top_p=invocation_params["top_p"],
            frequency_penalty=invocation_params["frequency_penalty"],
            presence_penalty=invocation_params["presence_penalty"],
        )
    elif invocation_params["_type"] == "openai-chat":
        return LLMSettings(
            model_name=invocation_params["model_name"],
            stop=invocation_params["stop"],
        )
    else:
        return None


class BaseChainlitCallbackHandler(BaseCallbackHandler):
    emitter: ChainlitEmitter
    # Keep track of the formatted prompts to display them in the prompt playground.
    prompts: List[str]
    # Keep track of the LLM settings for the last prompt
    llm_settings: LLMSettings
    # Keep track of the call sequence, like [AgentExecutor, LLMMathChain, Calculator, ...]
    sequence: List[str]
    # Keep track of the last prompt for each session
    last_prompt: Union[str, None]
    # Keep track of the currently streamed message for the session
    stream: Union[Message, None]

    raise_error = True

    # We want to handler to be called on every message
    always_verbose: bool = True

    def __init__(self) -> None:
        self.emitter = get_emitter()
        self.prompts = []
        self.llm_settings = None
        self.sequence = []
        self.last_prompt = None
        self.stream = None

    def end_stream(self):
        self.stream = None

    def add_in_sequence(self, name: str):
        self.sequence.append(name)

    def pop_sequence(self):
        if self.sequence:
            return self.sequence.pop()

    def add_prompt(self, prompt: str, llm_settings: LLMSettings = None):
        self.prompts.append(prompt)
        self.llm_settings = llm_settings

    def pop_prompt(self):
        if self.prompts:
            self.last_prompt = self.prompts.pop()

    def consume_last_prompt(self):
        last_prompt = self.last_prompt
        self.last_prompt = None
        return last_prompt

    def get_message_params(self):
        llm_settings = self.llm_settings

        indent = len(self.sequence) if self.sequence else 0

        if self.sequence:
            author = self.sequence[-1]
        else:
            author = config.ui.name

        return author, indent, llm_settings


class ChainlitCallbackHandler(BaseChainlitCallbackHandler, BaseCallbackHandler):
    def start_stream(self):
        author, indent, llm_settings = self.get_message_params()

        if author in IGNORE_LIST:
            return

        if config.code.lc_rename:
            author = run_sync(
                config.code.lc_rename(author, __chainlit_emitter__=self.emitter)
            )

        self.pop_prompt()

        __chainlit_emitter__ = self.emitter

        streamed_message = Message(
            author=author,
            indent=indent,
            llm_settings=llm_settings,
            prompt=self.consume_last_prompt(),
            content="",
        )
        self.stream = streamed_message

    def send_token(self, token: str):
        if self.stream:
            run_sync(self.stream.stream_token(token))

    def add_message(self, message, prompt: str = None, error=False):
        author, indent, llm_settings = self.get_message_params()

        if author in IGNORE_LIST:
            return

        if config.code.lc_rename:
            author = run_sync(
                config.code.lc_rename(author, __chainlit_emitter__=self.emitter)
            )

        __chainlit_emitter__ = self.emitter

        if error:
            run_sync(ErrorMessage(author=author, content=message).send())
            self.end_stream()
            return

        if self.stream:
            run_sync(self.stream.send())
            self.end_stream()
        else:
            run_sync(
                Message(
                    author=author,
                    content=message,
                    indent=indent,
                    prompt=prompt,
                    llm_settings=llm_settings,
                ).send()
            )

    # Callbacks for various events

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        invocation_params = kwargs.get("invocation_params")
        llm_settings = get_llm_settings(invocation_params)
        self.add_prompt(prompts[0], llm_settings)

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        **kwargs: Any,
    ) -> None:
        invocation_params = kwargs.get("invocation_params")
        llm_settings = get_llm_settings(invocation_params)
        prompt = "\n".join([m.content for m in messages[0]])
        self.add_prompt(prompt, llm_settings)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        if not self.stream:
            self.start_stream()
        self.send_token(token)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        self.pop_prompt()
        if response.llm_output is not None:
            if "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                if "total_tokens" in token_usage:
                    run_sync(
                        self.emitter.update_token_count(token_usage["total_tokens"])
                    )

    def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        pass

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        self.add_in_sequence(serialized["id"][-1])
        # Useful to display details button in the UI
        self.add_message("")

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        output_key = list(outputs.keys())[0]
        if output_key:
            prompt = self.consume_last_prompt()
            self.add_message(outputs[output_key], prompt)
        self.pop_sequence()

    def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        if isinstance(error, InterruptedError):
            return
        self.add_message(str(error), error=True)
        self.pop_sequence()

    def on_tool_start(
        self, serialized: Dict[str, Any], inputs: Any, **kwargs: Any
    ) -> None:
        self.add_in_sequence(serialized["name"])
        self.add_message("")

    def on_tool_end(
        self,
        output: str,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        prompt = self.consume_last_prompt()
        self.add_message(output, prompt)
        self.pop_sequence()

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Do nothing."""
        if isinstance(error, InterruptedError):
            return
        self.add_message(str(error), error=True)
        self.pop_sequence()

    def on_text(self, text: str, **kwargs: Any) -> None:
        pass

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        pass

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Run on agent end."""
        pass


class AsyncChainlitCallbackHandler(BaseChainlitCallbackHandler, AsyncCallbackHandler):
    async def start_stream(self):
        author, indent, llm_settings = self.get_message_params()

        if author in IGNORE_LIST:
            return

        if config.code.lc_rename:
            author = await config.code.lc_rename(
                author, __chainlit_emitter__=self.emitter
            )

        self.pop_prompt()

        __chainlit_emitter__ = self.emitter

        streamed_message = Message(
            author=author,
            indent=indent,
            prompt=self.consume_last_prompt(),
            llm_settings=llm_settings,
            content="",
        )
        self.stream = streamed_message

    async def send_token(self, token: str):
        if self.stream:
            await self.stream.stream_token(token)

    async def add_message(self, message, prompt: str = None, error=False):
        author, indent, llm_settings = self.get_message_params()

        if author in IGNORE_LIST:
            return

        if config.code.lc_rename:
            author = await config.code.lc_rename(
                author, __chainlit_emitter__=self.emitter
            )

        __chainlit_emitter__ = self.emitter

        if error:
            await ErrorMessage(author=author, content=message).send()
            self.end_stream()
            return

        if self.stream:
            await self.stream.send()
            self.end_stream()
        else:
            await Message(
                author=author,
                content=message,
                indent=indent,
                prompt=prompt,
                llm_settings=llm_settings,
            ).send()

    # Callbacks for various events

    async def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        invocation_params = kwargs.get("invocation_params")
        llm_settings = get_llm_settings(invocation_params)
        self.add_prompt(prompts[0], llm_settings)

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        **kwargs: Any,
    ) -> None:
        invocation_params = kwargs.get("invocation_params")
        llm_settings = get_llm_settings(invocation_params)
        prompt = "\n".join([m.content for m in messages[0]])
        self.add_prompt(prompt, llm_settings)

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        if not self.stream:
            await self.start_stream()
        await self.send_token(token)

    async def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        self.pop_prompt()
        if response.llm_output is not None:
            if "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                if "total_tokens" in token_usage:
                    await self.emitter.update_token_count(token_usage["total_tokens"])

    async def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        pass

    async def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        self.add_in_sequence(serialized["id"][-1])
        # Useful to display details button in the UI
        await self.add_message("")

    async def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        output_key = list(outputs.keys())[0]
        if output_key:
            prompt = self.consume_last_prompt()
            await self.add_message(outputs[output_key], prompt)
        self.pop_sequence()

    async def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        if isinstance(error, InterruptedError):
            return
        await self.add_message(str(error), error=True)
        self.pop_sequence()

    async def on_tool_start(
        self, serialized: Dict[str, Any], inputs: Any, **kwargs: Any
    ) -> None:
        self.add_in_sequence(serialized["name"])
        await self.add_message("")

    async def on_tool_end(
        self,
        output: str,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        prompt = self.consume_last_prompt()
        await self.add_message(output, prompt)
        self.pop_sequence()

    async def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Do nothing."""
        if isinstance(error, InterruptedError):
            return
        await self.add_message(str(error), error=True)
        self.pop_sequence()

    async def on_text(self, text: str, **kwargs: Any) -> None:
        pass

    async def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        pass

    async def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Run on agent end."""
        pass
